package de.jottyfan.timetrack.modules.done;

import static de.jottyfan.timetrack.db.done.Tables.T_DONE;
import static de.jottyfan.timetrack.db.done.Tables.T_JOB;
import static de.jottyfan.timetrack.db.done.Tables.T_MODULE;
import static de.jottyfan.timetrack.db.done.Tables.T_PROJECT;
import static de.jottyfan.timetrack.db.done.Tables.V_HAMSTERSUMMARY;
import static de.jottyfan.timetrack.db.done.Tables.V_TOTALOFDAY;
import static de.jottyfan.timetrack.db.done.Tables.V_WORKTIME;
import static de.jottyfan.timetrack.db.profile.Tables.T_LOGIN;

import java.sql.SQLException;
import java.text.SimpleDateFormat;
import java.time.LocalDate;
import java.time.LocalDateTime;
import java.time.ZoneId;
import java.time.format.DateTimeFormatter;
import java.util.ArrayList;
import java.util.Date;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.jooq.CloseableDSLContext;
import org.jooq.DeleteConditionStep;
import org.jooq.InsertValuesStep7;
import org.jooq.Record;
import org.jooq.Record3;
import org.jooq.Record4;
import org.jooq.Record5;
import org.jooq.Record6;
import org.jooq.SelectConditionStep;
import org.jooq.SelectJoinStep;
import org.jooq.SelectWhereStep;
import org.jooq.UpdateConditionStep;
import org.jooq.exception.DataAccessException;

import de.jooqfaces.JooqFacesContext;
import de.jottyfan.timetrack.db.done.tables.records.TDoneRecord;
import de.jottyfan.timetrack.db.done.tables.records.TJobRecord;
import de.jottyfan.timetrack.db.done.tables.records.TModuleRecord;
import de.jottyfan.timetrack.db.done.tables.records.TProjectRecord;
import de.jottyfan.timetrack.modules.JooqGateway;
import net.bootsfaces.component.fullCalendar.FullCalendarEventBean;
import net.bootsfaces.component.fullCalendar.FullCalendarEventList;

/**
 * 
 * @author henkej
 *
 */
public class DoneGateway extends JooqGateway {
	private final static Logger LOGGER = LogManager.getLogger(DoneGateway.class);
	private final Integer requested;

	public DoneGateway(JooqFacesContext facesContext) {
		super(facesContext);
		String dailyMinutes = facesContext.getExternalContext().getInitParameter("requestedDailyMinutes");
		this.requested = dailyMinutes == null ? 480 : Integer.valueOf(dailyMinutes); // defaults to 8 hours/day
	}

	/**
	 * get all modules from db
	 * 
	 * @return modules
	 * @throws SQLException
	 * @throws ClassNotFoundException
	 */
	public List<TModuleRecord> getAllModules() throws DataAccessException, ClassNotFoundException, SQLException {
		try (CloseableDSLContext jooq = getJooq()) {
			List<TModuleRecord> list = new ArrayList<>();
			SelectWhereStep<TModuleRecord> sql = jooq.selectFrom(T_MODULE);
			LOGGER.debug(sql.toString());
			for (TModuleRecord r : sql.fetch()) {
				list.add(r);
			}
			list.sort((o1, o2) -> o1 == null || o2 == null || o1.getName() == null || o2.getName() == null ? 0
					: o1.getName().compareTo(o2.getName()));
			return list;
		}
	}

	/**
	 * get all activities from db
	 * 
	 * @return activities
	 * @throws SQLException
	 * @throws ClassNotFoundException
	 */
	public List<TJobRecord> getAllActivities() throws DataAccessException, ClassNotFoundException, SQLException {
		try (CloseableDSLContext jooq = getJooq()) {
			List<TJobRecord> list = new ArrayList<>();
			SelectWhereStep<TJobRecord> sql = jooq.selectFrom(T_JOB);
			LOGGER.debug(sql.toString());
			for (TJobRecord r : sql.fetch()) {
				list.add(r);
			}
			list.sort((o1, o2) -> o1 == null || o2 == null || o1.getName() == null || o2.getName() == null ? 0
					: o1.getName().compareTo(o2.getName()));
			return list;
		}
	}

	/**
	 * get all projects from db
	 * 
	 * @return projects
	 * @throws SQLException
	 * @throws ClassNotFoundException
	 */
	public List<TProjectRecord> getAllProjects() throws DataAccessException, ClassNotFoundException, SQLException {
		try (CloseableDSLContext jooq = getJooq()) {
			List<TProjectRecord> list = new ArrayList<>();
			SelectWhereStep<TProjectRecord> sql = jooq.selectFrom(T_PROJECT);
			LOGGER.debug(sql.toString());
			for (TProjectRecord r : sql.fetch()) {
				list.add(r);
			}
			list.sort((o1, o2) -> o1 == null || o2 == null || o1.getName() == null || o2.getName() == null ? 0
					: o1.getName().compareTo(o2.getName()));
			return list;
		}
	}

	private Map<Integer, TProjectRecord> generateProjectMap(List<TProjectRecord> list) {
		Map<Integer, TProjectRecord> map = new HashMap<>();
		for (TProjectRecord r : list) {
			map.put(r.getPk(), r);
		}
		return map;
	}

	private Map<Integer, TModuleRecord> generateModuleMap(List<TModuleRecord> list) {
		Map<Integer, TModuleRecord> map = new HashMap<>();
		for (TModuleRecord r : list) {
			map.put(r.getPk(), r);
		}
		return map;
	}

	private Map<Integer, TJobRecord> generateJobMap(List<TJobRecord> list) {
		Map<Integer, TJobRecord> map = new HashMap<>();
		for (TJobRecord r : list) {
			map.put(r.getPk(), r);
		}
		return map;
	}

	/**
	 * get all from t_done of the given day
	 * 
	 * @param day
	 *          the day; if null, the current date is used
	 * 
	 * @return a list of found times, an empty one at least
	 * @throws SQLException
	 * @throws ClassNotFoundException
	 */
	public List<DoneBean> getAll(LocalDateTime day) throws DataAccessException, ClassNotFoundException, SQLException {
		Map<Integer, TProjectRecord> projectMap = generateProjectMap(getAllProjects());
		Map<Integer, TModuleRecord> moduleMap = generateModuleMap(getAllModules());
		Map<Integer, TJobRecord> jobMap = generateJobMap(getAllActivities());

		if (day == null) {
			day = LocalDateTime.now();
		}
		LocalDateTime yesterday = day.minusDays(1).withHour(23).withMinute(0).withSecond(0).withNano(0);
		LocalDateTime tomorrow = day.plusDays(1).withHour(0).withMinute(0).withSecond(0).withNano(0);

		List<DoneBean> list = new ArrayList<>();
		try (CloseableDSLContext jooq = getJooq()) {
			SelectConditionStep<TDoneRecord> sql = getJooq()
			// @formatter:off
				.selectFrom(T_DONE)
				.where(T_DONE.FK_LOGIN.eq(getFkLogin()))
				.and(T_DONE.TIME_FROM.isNull()
					.or(T_DONE.TIME_FROM.greaterThan(yesterday)
						.and(T_DONE.TIME_FROM.lessThan(tomorrow))))
				.and(T_DONE.TIME_UNTIL.isNull()
					.or(T_DONE.TIME_UNTIL.lessThan(tomorrow)
						.and(T_DONE.TIME_UNTIL.greaterThan(yesterday))));
			// @formatter:on
			LOGGER.debug(sql.toString());
			for (TDoneRecord r : sql.fetch()) {
				list.add(new DoneBean(r, projectMap, moduleMap, jobMap));
			}
		}
		list.sort((o1, o2) -> o1 == null || o2 == null ? 0 : o1.compareTo(o2));
		return list;
	}

	/**
	 * insert data into t_done
	 * 
	 * @param bean
	 * @throws SQLException
	 * @throws ClassNotFoundException
	 */
	public void insert(DoneBean bean) throws DataAccessException, ClassNotFoundException, SQLException {
		Integer fkProject = bean.getProject() == null ? null : bean.getProject().getPk();
		Integer fkModule = bean.getModule() == null ? null : bean.getModule().getPk();
		Integer fkJob = bean.getActivity() == null ? null : bean.getActivity().getPk();
		Integer fkLogin = getFkLogin();

		try (CloseableDSLContext jooq = getJooq()) {
			InsertValuesStep7<TDoneRecord, LocalDateTime, LocalDateTime, Integer, Integer, Integer, String, Integer> sql = jooq
			// @formatter:off
				.insertInto(T_DONE, 
										T_DONE.TIME_FROM, 
										T_DONE.TIME_UNTIL, 
										T_DONE.FK_PROJECT, 
										T_DONE.FK_MODULE, 
										T_DONE.FK_JOB,
										T_DONE.WP,
										T_DONE.FK_LOGIN)
				.values(bean.getTimeFrom(), bean.getTimeUntil(), fkProject, fkModule, fkJob, bean.getWp(), fkLogin);
			// @formatter:on
			LOGGER.debug(sql.toString());
			sql.execute();
		}
	}

	/**
	 * update bean in t_done
	 * 
	 * @param bean
	 * @throws SQLException
	 * @throws ClassNotFoundException
	 */
	public void update(DoneBean bean) throws DataAccessException, ClassNotFoundException, SQLException {
		try (CloseableDSLContext jooq = getJooq()) {
			UpdateConditionStep<TDoneRecord> sql = jooq
			// @formatter:off
				.update(T_DONE)
				.set(T_DONE.TIME_FROM, bean.getTimeFrom())
				.set(T_DONE.TIME_UNTIL, bean.getTimeUntil())
				.set(T_DONE.FK_PROJECT, bean.getProject() == null ? null : bean.getProject().getPk())
				.set(T_DONE.FK_JOB, bean.getActivity() == null ? null : bean.getActivity().getPk())
				.set(T_DONE.FK_MODULE, bean.getModule() == null ? null : bean.getModule().getPk())
				.set(T_DONE.WP, bean.getWp())
				.where(T_DONE.PK.eq(bean.getPk()));
			// @formatter:on
			LOGGER.debug(sql.toString());
			sql.execute();
		}
	}

	/**
	 * delete bean from t_done
	 * 
	 * @param bean
	 * @throws SQLException
	 * @throws ClassNotFoundException
	 */
	public void delete(DoneBean bean) throws DataAccessException, ClassNotFoundException, SQLException {
		try (CloseableDSLContext jooq = getJooq()) {
			DeleteConditionStep<TDoneRecord> sql = jooq.deleteFrom(T_DONE).where(T_DONE.PK.eq(bean.getPk()));
			LOGGER.debug(sql.toString());
			sql.execute();
		}
	}

	/**
	 * get day summary
	 * 
	 * @param day
	 *          the day
	 * @return the day summary if found, an empty map otherwise
	 * @throws SQLException
	 * @throws ClassNotFoundException
	 * @throws DataAccessException
	 */
	public WholeDaySummaryBean getDaySummary(java.util.Date day) throws DataAccessException, ClassNotFoundException, SQLException {
		try (CloseableDSLContext jooq = getJooq()) {
			SelectConditionStep<Record4<String, String, String, String>> sql = jooq
			// @formatter:off
				.select(V_TOTALOFDAY.STARTTIME,
								V_TOTALOFDAY.ENDTIME,
						    V_TOTALOFDAY.WORKTIME,
						    V_TOTALOFDAY.BREAKTIME)
				.from(V_TOTALOFDAY)
				.where(V_TOTALOFDAY.DAY.eq(new SimpleDateFormat("yyyy-MM-dd").format(day)))
				.and(V_TOTALOFDAY.FK_LOGIN.eq(getFkLogin()));
			// @formatter:on
			LOGGER.debug(sql.toString());
			Record r = sql.fetchOne();
			if (r != null) {
				String startTime = r.get(V_TOTALOFDAY.STARTTIME);
				String endTime = r.get(V_TOTALOFDAY.ENDTIME);
				String workTime = r.get(V_TOTALOFDAY.WORKTIME);
				String breakTime = r.get(V_TOTALOFDAY.BREAKTIME);
				return new WholeDaySummaryBean(startTime, endTime, workTime, breakTime, requested);
			}
			return new WholeDaySummaryBean("", "", "", "", requested);
		}
	}

	/**
	 * get all jobs of day
	 * 
	 * @param day
	 *          the day
	 * @return list of found jobs; an empty list at least
	 * @throws SQLException
	 * @throws ClassNotFoundException
	 * @throws DataAccessException
	 */
	public List<DailySummaryBean> getAllJobs(java.util.Date day) throws DataAccessException, ClassNotFoundException, SQLException {
		try (CloseableDSLContext jooq = getJooq()) {
			SelectConditionStep<Record6<String, Double, String, String, String, String>> sql = jooq
			// @formatter:off
				.select(V_WORKTIME.DURATION,
				        V_WORKTIME.DURATION_HOURS,
				        V_WORKTIME.PROJECT_NAME,
				        V_WORKTIME.MODULE_NAME,
				        V_WORKTIME.JOB_NAME,
				        V_WORKTIME.WP)
				.from(V_WORKTIME)
				.where(V_WORKTIME.DAY.eq(new SimpleDateFormat("yyyy-MM-dd").format(day)))
				.and(V_WORKTIME.FK_LOGIN.eq(getFkLogin()));
			// @formatter:on
			LOGGER.debug(sql.toString());
			List<DailySummaryBean> list = new ArrayList<>();
			for (Record6<String, Double, String, String, String, String> r : sql.fetch()) {
				String duration = r.get(V_WORKTIME.DURATION);
				Double durationHours = r.get(V_WORKTIME.DURATION_HOURS);
				String projectName = r.get(V_WORKTIME.PROJECT_NAME);
				String moduleName = r.get(V_WORKTIME.MODULE_NAME);
				String jobName = r.get(V_WORKTIME.JOB_NAME);
				String wp = r.get(V_WORKTIME.WP);
				list.add(new DailySummaryBean(projectName, moduleName, jobName, duration, wp, durationHours));
			}
			return list;
		}
	}

	/**
	 * get json representation of all calendar events of user
	 * 
	 * @return
	 * @throws SQLException
	 * @throws ClassNotFoundException
	 * @throws DataAccessException
	 */
	public String getAllCalendarEvents() throws DataAccessException, ClassNotFoundException, SQLException {
		try (CloseableDSLContext jooq = getJooq()) {
			SelectConditionStep<Record6<LocalDateTime, LocalDateTime, String, String, String, String>> sql = jooq
			// @formatter:off
				.select(T_DONE.TIME_FROM,
								T_DONE.TIME_UNTIL,
								T_DONE.WP,
								T_PROJECT.NAME,
								T_MODULE.NAME,
								T_JOB.NAME)
				.from(T_DONE)
				.leftJoin(T_PROJECT).on(T_PROJECT.PK.eq(T_DONE.FK_PROJECT))
				.leftJoin(T_MODULE).on(T_MODULE.PK.eq(T_DONE.FK_MODULE))
				.leftJoin(T_JOB).on(T_JOB.PK.eq(T_DONE.FK_JOB))
				.where(T_DONE.FK_LOGIN.eq(getFkLogin()));
			// @formatter:on
			LOGGER.debug(sql.toString());
			FullCalendarEventList list = new FullCalendarEventList();
			for (Record r : sql.fetch()) {
				String projectName = r.get(T_PROJECT.NAME);
				String moduleName = r.get(T_MODULE.NAME);
				String jobName = r.get(T_JOB.NAME);
				String wp = r.get(T_DONE.WP);
				LocalDateTime timeFrom = r.get(T_DONE.TIME_FROM);
				LocalDateTime timeUntil = r.get(T_DONE.TIME_UNTIL);

				StringBuilder buf = new StringBuilder();
				buf.append(projectName);
				if (wp != null && !wp.isBlank()) {
				  buf.append(" (").append(wp).append(")");
				}
				buf.append(", ");
				buf.append(moduleName);
				buf.append(": ");
				buf.append(jobName);

				FullCalendarEventBean bean = new FullCalendarEventBean(buf.toString(), java.util.Date.from(timeFrom.atZone(ZoneId.systemDefault()).toInstant())) {
					private static final long serialVersionUID = 1L;

					@Override
					public void addExtendedFields(StringBuilder buf) {
					}
				};
				Date endDate = timeUntil == null ? null : Date.from(timeUntil.atZone(ZoneId.systemDefault()).toInstant());
				bean.setEnd(endDate);
				bean.setColor(new RgbColor().determineRgbColor(projectName, moduleName, jobName));
				list.getList().add(bean);
			}
			return list.toJson();
		}
	}

	public List<UserBean> getAllUsers() throws DataAccessException, ClassNotFoundException, SQLException {
		List<UserBean> list = new ArrayList<>();
		try (CloseableDSLContext jooq = getJooq()) {
			SelectJoinStep<Record3<String, String, String>> sql = jooq
			// @formatter:off
				.select(T_LOGIN.LOGIN,
								T_LOGIN.FORENAME,
								T_LOGIN.SURNAME)
				.from(T_LOGIN);
			// @formatter:on
			LOGGER.debug(sql.toString());
			for (Record r : sql.fetch()) {
				String username = r.get(T_LOGIN.LOGIN);
				String forename = r.get(T_LOGIN.FORENAME);
				String surname = r.get(T_LOGIN.SURNAME);
				list.add(new UserBean(username, forename, surname));
			}
		}
		return list;
	}

	public String download(DownloadBean bean) throws DataAccessException, ClassNotFoundException, SQLException {
		StringBuilder buf = new StringBuilder();
		try (CloseableDSLContext jooq = getJooq()) {
			SelectConditionStep<Record5<LocalDate, String, String, String, String>> sql = jooq
			// @formatter:off
				.select(V_HAMSTERSUMMARY.WORKDAY,
						    V_HAMSTERSUMMARY.DURATION,
						    V_HAMSTERSUMMARY.PROJECT_NAME,
						    V_HAMSTERSUMMARY.MODULE_NAME,
						    V_HAMSTERSUMMARY.JOB_NAME)
				.from(V_HAMSTERSUMMARY)
				.where(V_HAMSTERSUMMARY.LOGIN.eq(bean.getUsername()))
				.and(V_HAMSTERSUMMARY.WORKDAY.le(bean.getUntilDate())
				.and(V_HAMSTERSUMMARY.WORKDAY.ge(bean.getFromDate())));
			// @formatter:on
			LOGGER.debug(sql.toString());
			String sep = ";";
			buf.append("day").append(sep).append("duration").append(sep).append("project").append(sep).append("module").append(sep).append("activity\n");
			for (Record r : sql.fetch()) {
				LocalDate workday = r.get(V_HAMSTERSUMMARY.WORKDAY);
				String date = workday.format(DateTimeFormatter.ofPattern("dd.MM.yyyy"));
				buf.append(date).append(sep);
				buf.append(r.get(V_HAMSTERSUMMARY.DURATION)).append(sep);
				buf.append(r.get(V_HAMSTERSUMMARY.PROJECT_NAME)).append(sep);
				buf.append(r.get(V_HAMSTERSUMMARY.MODULE_NAME)).append(sep);
				buf.append(r.get(V_HAMSTERSUMMARY.JOB_NAME)).append("\n");
			}
		}
		return buf.toString();
	}
}
